local Block = require("blocks.Block")
local EventDispatcher = require("events.EventDispatcher")
local Noise = require("generation.noise.Noise")
local materials = require("material")
local blocks = require("blocks.blocks")
local points = require("point")
local Point = points.Point
local deltas = points.deltas
local debugUtil = require("debug.debugUtil")
local Generator = require("generation.generators.Generator")
local Chunk = require("game.Chunk")
local LiquidHelper = require("physics.LiquidHelper")


local function make2dArray(width,height, constructor)  
  local result = {}
  for y=1,height do
    local row = {}
    result[y] = row
    for x=1,width do
      row[x] = constructor(x,y)
      if row[x]==nil then
        error("constructor was faulty!")
      end
    end
  end  
  return result
end








local BlockLayer = {}
function BlockLayer:new(width, height, seed, index) 
  self.__index = self
  if not index then
    index = 0
  end
  local instance = { 
    width=width, 
    height=height, 
    blocks = {},
    groundBlockLayer=math.ceil(height/2),
    toUpdate={},
    manyUpdatesCheck = 0,
    generator = Generator:new(seed),
    chunkSize = 16,
    chunkLoaders = {},
    chunks = {},
    index = index
  }
  
  setmetatable(instance, self)

  instance.generator.groundBlockLayer = instance.groundBlockLayer
  instance.generator.height = instance.height
  
  instance.liquidHelper = LiquidHelper:new(instance)
  
  instance.defaultBlock = blocks.new("core:grass")
  instance.defaultBlock.destructible = false
  --instance:generateBlocks()
  EventDispatcher:new(instance)
 
  return instance
end

function BlockLayer:handleChunkLoaders()  
  local start = love.timer.getTime()
  local toLoad = {}
  local count = 0

  for _, chunkLoader in ipairs(self.chunkLoaders) do
    local pos = chunkLoader.position:clone()
    if not chunkLoader.isBlock then
      --print("pos "..tostring(pos).. "becomes... ")
      pos = (pos * (1/BLOCK_SIZE)):round()
      --print("    ..."..tostring(pos))
    end
    local r = 4
    for dx=-r,r do
      for dy=-r,r do
        local chunk =  self:chunkAtBlockXY(pos.x+dx*self.chunkSize,pos.y+dy*self.chunkSize)
        toLoad[chunk] = true
      end
    end      
  end
  
  local toUnload = {}
  for key, chunk in pairs(self.chunks) do
    if not toLoad[chunk] then
      self:clearChunk(chunk)                    
      count = count + 1      
    end
  end
  
  for chunk, _ in pairs(toLoad) do          
    if not chunk.isLoaded then               
      chunk:load()
      count = count + 1            
    end
  end
  
  if count > 0 then
    local delta = (love.timer.getTime() - start)
    if delta > 0.1 then
      print("chunk operations = "..tostring(count).." times "..delta)    
    end
  end

 
  
  
end

function BlockLayer:chunkAtBlockXY(x,y)
  x = x - fmod(x, self.chunkSize)
  assert(fmod(y, self.chunkSize) >= 0)
  y = y - fmod(y, self.chunkSize)
  
  local key = x..","..y
  
  if not self.chunks[key] then
    local chunk = Chunk:new(self, Point:new(x,y), Point:new(x+self.chunkSize-1, y+self.chunkSize-1))
    self.chunks[key]= chunk    
  end
  return self.chunks[key]
end

function BlockLayer:chunkAtEntityPos(pt)
  local pos = pt:clone()
  pos = (pos * (1/BLOCK_SIZE)):round()
  return self:chunkAtBlockXY(pos.x, pos.y)
end

function BlockLayer:clearChunk(chunk)
  local key = chunk.fromPoint.x..","..chunk.fromPoint.y
  if chunk.isLoaded then
    chunk:unload()
  end
  self.chunks[key] = nil
end


function BlockLayer:generateBlocks()
  local groundBlockLayer = self.groundBlockLayer;
  local width = self.width
  local height = self.height
  print("groundlevel="..groundBlockLayer)
  self.blocks = make2dArray(width, height, function(x,y) return false end)
  
  
  local generator = self.generator
  for x=1, width do       
    for y =1, height do
      self:setBlock(x, y, generator:getBlockXY(x,y))      
    end
  end
end

function BlockLayer:getBlock(x,y)
  local row = self.blocks[y]
  local val = row and row[x] or nil
  return val or self.defaultBlock  
  --[[return (self.blocks[y] and self.blocks[y][x]) and self.blocks[y][x]
  local blocks = self.blocks
  if blocks[y] and blocks[y][x] then
    return blocks[y][x]
  else
    return self.defaultBlock:clone()
  end ]]-- 
end

function BlockLayer:getBlockPt(pt)
  return self:getBlock(pt.x, pt.y)
end


function BlockLayer:clearBlock(x,y)
  local _blocks = blocks
  local blocks = self.blocks
  if blocks[y] then 
    local block = blocks[y][x]
    if block then
      block:setLevel(nil)
      _blocks.free(block)
    end
    blocks[y][x] = nil
  end  
  local good=false
  for _, _ in pairs(blocks[y]) do
    good = true
    break
  end
  if not good then
    blocks[y] = nil
  end
end

function BlockLayer:spawnBlock(x,y, block)
  if not block then
    return
  end
  if not self.blocks[y] then
    self.blocks[y]= {}
  end
  self.blocks[y][x] = block  
  block:setLevel(self)
  block.position.x = x
  block.position.y = y

end

function BlockLayer:setBlock(x,y, block)
  local _blocks = blocks
  local blocks = self.blocks
  local function assertInt(value)
    assert(type(value)=="number" and math.floor(value)==value, ""..value.." not int")
    return value
  end
  assertInt(x)
  assertInt(y)  
  assert(block, "setting nil at "..x..","..y)
  local oldLevel = block:getLevel()
  if oldLevel then
    oldLevel:setBlock(block.position.x, block.position.y, _blocks.new("core:sky"))
  end
  
  if blocks[y] and blocks[y][x] then
    local oldBlock = blocks[y][x]
    oldBlock:setLevel(nil)
  end
  if not blocks[y] then
    print("fail at "..x.." "..y.." ")
  end
  blocks[y][x] = block
  block.position.x = x
  block.position.y = y
  
  block:setLevel(self)
  
  self:scheduleNeighboursForUpdate(block)
  
  
  return block    
end

function BlockLayer:setBlockPt(pos, block)
  self:setBlock(pos.x, pos.y, block)
end

function BlockLayer:scheduleNeighboursForUpdate(block, radius)
  if not radius then
    radius = 1
  end
  local centerX = block.position.x
  local centerY = block.position.y
  for dx=-radius, radius do
    for dy=-radius, radius do
      self.toUpdate[self:getBlock(centerX+dx, centerY+dy)] = true
    end
  end
end




function BlockLayer:updateBlocks()
  local toUpdate = self.toUpdate
  self.toUpdate = {}
  local priorityKeys = {}
  local priorities = {}
  for block, _ in pairs(toUpdate) do
    local priority = block.updatePriority and block.updatePriority or 0
    if not priorities[priority] then
      priorities[priority] = {}
      table.insert(priorityKeys, priority)
    end
    table.insert(priorities[priority], block)
  end
  
  table.sort(priorityKeys, function(a,b) return a > b end)
  for _, priority in ipairs(priorityKeys) do
    for __, block in ipairs(priorities[priority]) do
      block:update()
    end
  end

  --[[self.manyUpdatesCheck = self.manyUpdatesCheck + 1
  for block, _ in pairs(toUpdate) do
    block:update()
  end]]--
  
end

function BlockLayer:update()
  self:handleChunkLoaders()
  self:updateBlocks()
end


function BlockLayer:dropBlock(block)  
  self:setBlock(block.position.x, block.position.y, blocks.new("core:sky"))
  self:dispatchEvent("dropBlock", block)
end



function BlockLayer:getBlockPx(pixelPos)  
  local blockX = math.floor(pixelPos.x / BLOCK_SIZE)
  local blockY = math.floor(pixelPos.y / BLOCK_SIZE)
  --print("xy "..blockX.." "..blockY)
  return self:getBlock(blockX, blockY)  
end

return BlockLayer


  