local Point = require("point").Point

local M = {}
local Renderer = {}
M.Renderer = Renderer

function Renderer:new(blockSize, screenWidth, screenHeight)
  self.__index = self
  local newObj = {
    blockSize = blockSize,
    screenWidth = screenWidth,
    screenHeight = screenHeight
    }
  return setmetatable(newObj, self)
end

function Renderer:setHud(hud)
  self.hud = hud
end
    

function Renderer:render(gameArea, centerPos)  
  centerPos = centerPos:round()
  self.centerPos = centerPos
  local screeHalf = Point:new(self.screenWidth/2, self.screenHeight/2):round():assertInt()
  
  self.topLeftShift = (centerPos - screeHalf):assertInt()
    
  for i, layer in ipairs(gameArea.layers) do    
    self:renderLayer(layer, i)        
  end
  
  for i, entity in ipairs(gameArea.entities) do    
    self:renderEntity(entity)
  end
  
  local hud = self.hud
  if hud then
    hud.size = Point:new(love.graphics.getWidth(), love.graphics.getHeight())
    hud.position:set2(0,0)
    self.hud:draw()
  end
  
      
end

function Renderer:renderEntity(entity)
  local pos = entity.position:round() - self.topLeftShift:assertInt()
  
  local topLeft = pos
  local bottomRight = topLeft + Point:new(entity.shapeWidth, entity.shapeHeight)
  
  if bottomRight.x < 0 or bottomRight.y < 0 or topLeft.x > self.screenWidth or topLeft.y > self.screenHeight then
    -- offscreen entity - no need for rendering
    return
  end
  
  local tex = textures[entity.material.textureId]
  assert(tex, "no texture id "..tostring(entity.material.textureId))
  local scaleX = entity.shapeWidth / tex:getWidth()
  local scaleY = entity.shapeHeight / tex:getHeight()
  assert(pos.x==math.floor(pos.x))
  assert(pos.y==math.floor(pos.y))
  love.graphics.draw(tex, pos.x, pos.y, nil, scaleX, scaleY)
  
  if entity.equipment then
    for equipment, v in pairs(entity.equipment) do
      local targetPos = pos + equipment.shift
      local size = equipment.targetSize
      local c = {love.graphics.getColor()}
      love.graphics.setColor(60,255,60,128)
      love.graphics.rectangle("fill", targetPos.x, targetPos.y, size.x, size.y)
      love.graphics.setColor(unpack(c))
    end
  end
end

-- iterator for (index, position) based on center and screen size
local function forBlockIndexAndPosition(center, length, blockSize)
  local startPixel = center - math.floor(length/2+0.5)
  local blockCount = math.ceil( length / blockSize)+1
  local startBlockIndex = math.floor(startPixel/blockSize)
  
  
  
  local currentIndexShift = 0
  return function()
    local i = currentIndexShift
    currentIndexShift  = currentIndexShift + 1
    
    if i >= blockCount then
      return nil
    else
      local blockIndex = startBlockIndex + i
      return blockIndex, blockIndex * blockSize - startPixel
    end
  end    
end   

local batches = {}

function Renderer:renderLayer(layer, renderIndex)
  if not self.atlas then
    self.atlasSize = 10
    self.atlas = love.image.newImageData(self.blockSize * self.atlasSize, self.blockSize * self.atlasSize)
    self.atlasBatch = love.graphics.newSpriteBatch(love.graphics.newImage(self.atlas), 4*math.floor( self.screenHeight*self.screenWidth/self.blockSize/self.blockSize), "stream")    
    self.textureIndices = {}    
    self.quads = {}
  end
  
  local batch = self.atlasBatch  
  
  local atlasUpdated = false
  local function getTextureQuad(texture)
    if not self.textureIndices.count then
      self.textureIndices.count = 0
    end
    local index = self.textureIndices[texture]
    local newTexture = not index
    if newTexture then
      index = self.textureIndices.count 
      self.textureIndices[texture]  = index
      self.textureIndices.count = index + 1
    end
    
    local dx = self.blockSize * (index % self.atlasSize)
    local dy = self.blockSize * math.floor(index/self.atlasSize)
    if newTexture then
      atlasUpdated = true
      self.atlas:paste(texture:getData(), dx, dy, 0, 0, self.blockSize, self.blockSize)
      self.quads[index] = love.graphics.newQuad(dx,dy, self.blockSize, self.blockSize, self.atlas:getDimensions())  
    end
   -- print("quad="..tostring(dx).." "..tostring(dy))
    return self.quads[index]
  end
  
  local level = layer
  local centerPos = self.centerPos  
 
  
  local blockCounter = 0
  local tempColor = {love.graphics.getColor()}
  
    
  
  for yI, yPos in forBlockIndexAndPosition(centerPos.y, self.screenHeight, self.blockSize) do  
    for xI, xPos in forBlockIndexAndPosition(centerPos.x, self.screenWidth, self.blockSize) do    
      local block = level:getBlock(xI,yI)
     
          
--      if not block.material then 
--        print("faulty");
--        for k,v in pairs(block) do print(k.."="..tostring(v)) end
--      end
--      assert(block.material, "no material id="..tostring(block.__index.blockId).." pos="..tostring(block.position))
      local texture = textures[block.material.textureId];
      --assert(texture, "no texture with id "..tostring(block.material.textureId))
      
      if not texture then
        texture = textures["broken"]
        --print("no texture with id "..tostring(block.material.textureId))
      end
            
      
      
      local durabilityRatio = block.durability / block.maxDurability
      
      
      if durabilityRatio > 0 then
        local isSky = block.blockId == "core:sky"
        local showSky = renderIndex == 1 and isSky
        local bright = layer.index == 0
        local brightness = (showSky) and 1 or (math.pow(0.6, layer.index))   
        local alpha = (isSky and not showSky) and 0 or 255
        
        local fullColor = brightness * 255
        local durabilityColor = fullColor * durabilityRatio
        
        batch:setColor(fullColor, durabilityColor, durabilityColor,alpha)
      else
        batch:setColor(64,0,0)
      end
      
      --love.graphics.draw(texture, xPos, yPos)
      batch:add(getTextureQuad(texture),xPos,yPos)
      
      --getTextureQuad(texture)

      
      --[[if block.pressure then
        local oldFont = love.graphics.getFont()        
        love.graphics.setFont(font7)
        love.graphics.setColor(0,0,0)
        love.graphics.print(string.sub(tostring(block.pressure),0,3),xPos+4,yPos+4)
        love.graphics.setFont(oldFont)
      end]]--
      
      
      blockCounter = blockCounter + 1          
    
    end
  end
  if atlasUpdated then
    batch:setTexture(love.graphics.newImage(self.atlas))
  end
  batch:flush()
  love.graphics.draw(batch,0,0)
  batch:clear()
  
  love.graphics.setColor(unpack(tempColor))
  self.blocksRendered = blockCounter
  self.atlasUpdated = atlasUpdated
end

  
return M